# -*- coding: utf-8 -*-

__title__ = 'CryptoDataHub'
__technical_name__ = __title__.lower()
__version__ = '0.12.6'
__description__ = 'Repository of cryptography-related data'
__author__ = 'Szilárd Pfeiffer'
__author_email__ = 'coroner@pfeifferszilard.hu'
__url__ = 'https://gitlab.com/coroner/' + __technical_name__
__license__ = 'MPL-2.0'
